// ----------------------------------------------------------------
// The contents of this file are distributed under the CC0 license.
// See http://creativecommons.org/publicdomain/zero/1.0/
// ----------------------------------------------------------------

#include "View.h"
#include "Model.h"
#include "error.h"
#include <SDL2/SDL_image.h>


View::View(Model& m, int w, int h, double obs_noise_dev)
: model(m), visualizing(true), obs_dev(obs_noise_dev)
{
	// Init SDL stuff
	if(SDL_Init(SDL_INIT_VIDEO) < 0)
		throw Ex("Unable to Init SDL: ", SDL_GetError());
	if(!SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "1"))
		throw Ex("Unable to Init hinting: ", SDL_GetError());
	m_pWindow = SDL_CreateWindow("Crane Simulator", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, w, h, SDL_WINDOW_SHOWN);
	if(m_pWindow == NULL)
		throw Ex("Unable to create SDL Window: ", SDL_GetError());
	m_pPrimarySurface = SDL_GetWindowSurface(m_pWindow);
	m_pRenderer = SDL_GetRenderer(m_pWindow);
    	if(m_pRenderer == NULL)
	m_pRenderer = SDL_GetRenderer(m_pWindow);
	if(m_pRenderer == NULL)
		m_pRenderer = SDL_CreateRenderer(m_pWindow, -1, SDL_RENDERER_ACCELERATED);
	if(m_pRenderer == NULL)
		throw Ex("Unable to create renderer");

	// Initialize image loading for PNGs
	if(!(IMG_Init(IMG_INIT_PNG) & IMG_INIT_PNG)) {
		throw Ex("Unable to init SDL_image: ", IMG_GetError());
	}
}

// virtual
View::~View()
{
	if(m_pRenderer)
		SDL_DestroyRenderer(m_pRenderer);
//	if(m_pWindow)
//		SDL_DestroyWindow(m_pWindow);
	IMG_Quit();
//	SDL_Quit();
}

void View::drawRect(int x, int y, int w, int h)
{
	SDL_RenderDrawLine(m_pRenderer, x, y, x + w, y);
	SDL_RenderDrawLine(m_pRenderer, x, y + h, x + w, y + h);
	SDL_RenderDrawLine(m_pRenderer, x, y, x, y + h);
	SDL_RenderDrawLine(m_pRenderer, x + w, y, x + w, y + h);
}

#define GRID_SCALE 20

size_t state_to_screen_x(double x)
{
	size_t v = 100 + (size_t)(GRID_SCALE * (10.0 * x + 10));
	return v;
}

size_t state_to_screen_y(double y)
{
	size_t v = 20 * GRID_SCALE - (size_t)(GRID_SCALE * (10.0 * y + 10));
	return v;
}

void View::update()
{
	// Clear the screen
	SDL_SetRenderDrawColor(m_pRenderer, 0xff, 0xff, 0xff, 0xff);
	SDL_RenderClear(m_pRenderer);

	// Draw the predicted observation
	Vec in(4);
	for(size_t y = 0; y < 48; y++)
	{
		for(size_t x = 0; x < 64; x++)
		{
			in[0] = (double)x / 64;
			in[1] = (double)y / 48;
			in[2] = model.state[0];
			in[3] = model.state[1];
			const Vec& pred = model.obs.forwardprop(in);
			int r = std::max(0, std::min(255, (int)((pred[0] + model.rand.normal() * obs_dev) * 256)));
			int g = std::max(0, std::min(255, (int)((pred[1] + model.rand.normal() * obs_dev) * 256)));
			int b = std::max(0, std::min(255, (int)((pred[2] + model.rand.normal() * obs_dev) * 256)));
			SDL_SetRenderDrawColor(m_pRenderer, r, g, b, 0xff);
			SDL_RenderDrawPoint(m_pRenderer, x, y);
		}
	}

	// Draw a grid
	SDL_SetRenderDrawColor(m_pRenderer, 0xb0, 0xb0, 0xb0, 0xff);
	for(size_t i = 0; i <= 20; i++)
	{
		size_t x = 100 + GRID_SCALE * i;
		size_t y = GRID_SCALE * (20 - i);
		SDL_RenderDrawLine(m_pRenderer, 100, y, 100 + 20 * GRID_SCALE, y); // horizontal line
		SDL_RenderDrawLine(m_pRenderer, x, 0, x, 20 * GRID_SCALE); // vertical line
	}

	// Draw the state
	SDL_SetRenderDrawColor(m_pRenderer, 0x00, 0x00, 0x80, 0xff);
	SDL_Rect r;
	r.x = state_to_screen_x(model.state[0]) - 2;
	r.y = state_to_screen_y(model.state[1]) - 2;
	r.w = 5;
	r.h = 5;
	SDL_RenderFillRect(m_pRenderer, &r);

	SDL_RenderPresent(m_pRenderer);
}

