// ----------------------------------------------------------------
// The contents of this file are distributed under the CC0 license.
// See http://creativecommons.org/publicdomain/zero/1.0/
// ----------------------------------------------------------------

#include "error.h"
#include "string.h"
#include "json.h"
#include "rand.h"
#include "matrix.h"
#include "neuralnet.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <exception>
#include <string>
#include <memory>
#include "Model.h"
#include "View.h"
#include "Controller.h"
#ifdef WINDOWS
#	include <io.h> // for "access"
#else
#	include <unistd.h>
#endif

using std::cout;
using std::cerr;
using std::string;
using std::vector;

void mili_sleep(unsigned int nMiliseconds)
{
#ifdef WINDOWS
	MSG aMsg;
	while(PeekMessage(&aMsg, NULL, WM_NULL, WM_NULL, PM_REMOVE))
	{
		TranslateMessage(&aMsg);
		DispatchMessage(&aMsg);
	}
	SleepEx(nMiliseconds, 1);
#else
	nMiliseconds ? usleep(nMiliseconds * 1024) : sched_yield();
#endif
}



class CraneSimulator
{
public:
	NeuralNet* pObs;
	NeuralNet* pTrans;
	Rand& rand;


	CraneSimulator(Rand& r)
	: pObs(nullptr), pTrans(nullptr), rand(r)
	{
	}

	~CraneSimulator()
	{
		delete(pObs);
		delete(pTrans);
	}

	void trainTransitionModel()
	{
		// Get the transition data
		Matrix stateData;
		stateData.loadARFF("states.arff");
		stateData *= (1.0 / 10);
		Matrix actionData;
		actionData.loadARFF("actions.arff");

		// Make a new model
		delete(pTrans);
		pTrans = new NeuralNet();
		pTrans->layers().push_back(new LayerLinear(6, 8));
		pTrans->layers().push_back(new LayerTanh(8));
		pTrans->layers().push_back(new LayerLinear(8, 2));
		pTrans->init_weights(rand);

		// Train the transition model
		double learning_rate = 0.002;
		Vec in(6);
		Vec target(2);
		for(size_t i = 0; i < 1e7; i++)
		{
			int index = rand.next(actionData.rows() - 1);
			in.fill(0.0);
			in[0] = stateData[index][0];
			in[1] = stateData[index][1];
			size_t action = (size_t)actionData[index][0];
			in[2 + action] = 1.0;
			target[0] = stateData[index + 1][0] - stateData[index][0];
			target[1] = stateData[index + 1][1] - stateData[index][1];
			pTrans->train_incremental(in, target, learning_rate);
		}
	}

	void trainObservationModel()
	{
		// Get the observations data
		Matrix stateData;
		stateData.loadARFF("states.arff");
		stateData *= (1.0 / 10);
		Matrix obsData;
		obsData.loadARFF("observations.arff");
		obsData *= (1.0 / 256);

		// Make a new model
		delete(pObs);
		pObs = new NeuralNet();
		pObs->layers().push_back(new LayerLinear(4, 12));
		pObs->layers().push_back(new LayerTanh(12));
		pObs->layers().push_back(new LayerLinear(12, 12));
		pObs->layers().push_back(new LayerTanh(12));
		pObs->layers().push_back(new LayerLinear(12, 3));
		pObs->layers().push_back(new LayerTanh(3));
		pObs->init_weights(rand);

		// Train the observation model
		cout << "Training the observation model...\n";
		double learning_rate = 0.03;
		Vec in(4);
		Vec target(3);
		for(size_t j = 0; j < 100; j++)
		{
			cout << "\r" << to_str(j) << "% ";
			cout.flush();
			for(size_t i = 0; i < 2e5; i++)
			{
				size_t frame = rand.next(obsData.rows());
				size_t x = rand.next(64);
				size_t y = rand.next(48);
				in[0] = (double)x / 64;
				in[1] = (double)y / 48;
				in[2] = stateData[frame][0];
				in[3] = stateData[frame][1];
				size_t pos = 3 * (64 * y + x);
				Vec& obsVec = obsData[frame];
				target[0] = obsVec[pos];
				target[1] = obsVec[pos + 1];
				target[2] = obsVec[pos + 2];

				pObs->scale_gradient(0.0);
				pObs->forwardprop(in);
				pObs->compute_output_layer_blame_terms(target);
				pObs->backprop();
				pObs->update_gradient(in);
				pObs->step(learning_rate);
			}
			learning_rate *= 0.98;
		}
	}

	void loadModels()
	{
		// Load or build the observation function
		if(access("obs.json", 0) == 0) // If the file "obs.json" exists...
		{
			// Load the existing model from the file
			cout << "Loading the observation model from file...\n";
			Json doc;
			doc.loadJson("obs.json");
			delete(pObs);
			pObs = new NeuralNet(*doc.root());
		}
		else
		{
			// Train a new observation model
			trainObservationModel();
			cout << "Saving it to file...\n";
			Json doc;
			JsonNode* pNode = pObs->marshal(doc);
			pNode->saveJson("obs.json");
		}

		// Load or build the transition function
		if(access("trans.json", 0) == 0) // If the file "trans.json" exists...
		{
			// Load the existing model from the file
			cout << "Loading the transition model from file...\n";
			Json doc;
			doc.loadJson("trans.json");
			delete(pTrans);
			pTrans = new NeuralNet(*doc.root());
		}
		else
		{
			// Train a new transition model
			trainTransitionModel();
			cout << "Saving it to file...\n";
			Json doc;
			JsonNode* pNode = pTrans->marshal(doc);
			pNode->saveJson("trans.json");
		}
	}
};



int main(int argc, char *argv[])
{
	if(argc < 3)
	{
		cout << "Usage:   ./sim [trans_noise_dev] [obs_noise_dev]\n";
		cout << "Example: ./sim 0.0 0.0\n";
		return 1;
	}

	int nRet = 0;
	try
	{
		Rand rand(1234);
		CraneSimulator sim(rand);
		sim.loadModels();
		bool keepRunning = true;
		double trans_noise_dev = atof(argv[1]);
		double obs_noise_dev = atof(argv[2]);
		Model m(*sim.pObs, *sim.pTrans);
		View v(m, 500, 500, obs_noise_dev);
		Controller c(m, v, &keepRunning, trans_noise_dev);
		while(keepRunning)
		{
			m.update();
			if(v.visualizing)
				v.update();
			mili_sleep(30);
			c.update();
		}
	}
	catch(const std::exception& e)
	{
		cerr << e.what() << "\n";
		nRet = 1;
	}

	return nRet;
}

